Package: automatedRecLin
Title: Record Linkage Based on an Entropy-Maximizing Classifier
Version: 1.1.1
Authors@R: 
  c(person(given = "Adam",
           family = "Struzik",
           role = c("aut", "cre"),
           email = "adastr5@st.amu.edu.pl",
           comment = c(ORCID = "0009-0002-2547-482X")),
    person(given = "Maciej",
           family = "Beręsewicz",
           role = c("aut", "ctb"),
           email = "maciej.beresewicz@ue.poznan.pl",
           comment = c(ORCID = "0000-0002-8281-4301")))
Description: The goal of 'automatedRecLin' is to perform record linkage (also known as entity resolution) in unsupervised or supervised settings. It compares pairs of records from two datasets using selected comparison functions to estimate the probability or density ratio between matched and non-matched records. Based on these estimates, it predicts a set of matches that maximizes entropy. For details see: Lee et al. (2022) <https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm>, Vo et al. (2023) <https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html>, Sugiyama et al. (2008) <doi:10.1007/s10463-008-0197-x>.
License: GPL-3
Encoding: UTF-8
URL: https://github.com/ncn-foreigners/automatedRecLin
BugReports: https://github.com/ncn-foreigners/automatedRecLin/issues
Imports: blocking, data.table, densityratio, FixedPoint, methods,
        nleqslv, purrr, reclin2, stats, utils
Suggests: knitr, rmarkdown, tinytest, xgboost
Depends: R (>= 4.1.0)
LazyData: true
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-21 06:30:50 UTC; adamstruzik
Author: Adam Struzik [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-2547-482X>),
  Maciej Beręsewicz [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-8281-4301>)
Maintainer: Adam Struzik <adastr5@st.amu.edu.pl>
Repository: CRAN
Date/Publication: 2026-05-21 07:10:02 UTC
Built: R 4.6.0; ; 2026-05-22 00:25:19 UTC; unix
