# Generated by roxygen2: do not edit by hand

S3method(predict,rec_lin_model)
S3method(print,comparison_vectors)
S3method(print,mec_blocking)
S3method(print,mec_rec_lin)
S3method(print,rec_lin_model)
S3method(print,rec_lin_predictions)
export(abs_distance)
export(comparison_vectors)
export(control_kliep)
export(custom_rec_lin_model)
export(jarowinkler_complement)
export(mec)
export(mec_blocking)
export(train_rec_lin)
import(data.table)
import(stats)
importFrom(FixedPoint,FixedPoint)
importFrom(blocking,blocking)
importFrom(densityratio,kliep)
importFrom(methods,is)
importFrom(nleqslv,nleqslv)
importFrom(purrr,partial)
importFrom(reclin2,cmp_identical)
importFrom(reclin2,cmp_jarowinkler)
importFrom(stats,dbinom)
importFrom(stats,dgamma)
importFrom(stats,predict)
importFrom(stats,runif)
importFrom(utils,head)
