% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/controls.R
\name{control_kliep}
\alias{control_kliep}
\title{Controls for the \link[densityratio:kliep]{kliep()} Function}
\usage{
control_kliep(scale = NULL, progressbar = FALSE, nfold = 2, ...)
}
\arguments{
\item{scale}{\code{"numerator"}, \code{"denominator"} or \code{NULL}, indicating whether to standardize each numeric variable according
to the numerator means and standard deviations, the denominator means and standard deviations,
or apply no standardization at all.}

\item{progressbar}{Logical indicating whether or not to display a progress bar.}

\item{nfold}{Number of cross-validation folds used in order to calculate the optimal sigma value (default is 2-fold cross-validation).}

\item{...}{Additional arguments.}
}
\value{
Returns a list with parameters.
}
\description{
Controls for the \link[densityratio:kliep]{kliep()} function used in the package.
}
\author{
Adam Struzik
}
